# Copyright 2022 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Hashing algorithms."""

from streaming.base.hashing.hashing import get_hash, get_hashes, is_hash

__all__ = ['get_hash', 'get_hashes', 'is_hash']
